/*
   program : recon2img

   $Source: /home/yee/C/RCS/recon2img.c,v $
   $Author: yee $
   $Date: 1999/02/10 16:39:47 $
   $Revision: 1.36 $

   This program converts an image file to a standard format used by
   standard image viewers.

   Parameters : see the documentation for ParseArg().

   Assumptions : none known

   Bugs :
     1) TIFF support code ONLY supports 8 bit images.
     2) memory leak currently at point marked "memory leak". bug in mdfree().
        freeing of array commented out.

   To Do :
     1) file compression not yet supported.
     2) add support for file formats other than TIFF.

   $Log: recon2img.c,v $
   Revision 1.36  1999/02/10 16:39:47  yee
   *** empty log message ***

   Revision 1.35  1999/01/25 03:45:08  yee
   *** empty log message ***

   Revision 1.34  1999/01/24 06:39:57  yee
   change default window/level.

   Revision 1.33  1999/01/16 14:16:55  yee
   *** empty log message ***

   Revision 1.31  1999/01/16 06:35:16  yee
   support level/window to determine image limits.

   Revision 1.30  1999/01/15 04:20:28  yee
   fixed transposition bug

   Revision 1.29  1998/11/27 04:18:25  yee
   fix mapping error between image space and TIFF which caused rotation error.

   Copyright 1997,1998,1999 by Conway Yee.  All rights reserved.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include <tiff.h>
#include <tiffio.h>

#include "style.h"
#include "mutil.h"
#include "util.h"
#include "codes.h"
#include "array.h"

#include "backproject.h"
#include "ImageHeader.h"
#include "recon.h"

char recon2img_c[] = "$Id: recon2img.c,v 1.36 1999/02/10 16:39:47 yee Exp $";

#define DEFAULT_FMINPIX 0.0    /* default dynamic range */
#define DEFAULT_FMAXPIX 1.0 

#define DEFAULT_BITSPERPIXEL 8
#define RECON_DEFAULT_JPEGQUALITY 75
#define RECON_DEFAULT_ROWSPERSTRIP 10
#define RECON_DEFAULT_LZWPREDICTOR unknown   /* this line is not used */

#define OUTPUT_IMAGE_INT char

enum{IMAGE_GIF=1,IMAGE_TIFF};   /* sets image file type */
enum{UNDEFINED,EXTREMA,CENTER}; /* used to set image limits */

typedef union OUTPTR {
  FILE *ptr;              /* output file pointer */
  TIFF *tiff;
} OUTPTR;

typedef struct ConvHeader {
  CODE ftype;        /* output file type */
  CODE ctype;        /* compression type */
                                  /* the float equivalents are expressed
                                     as a fraction of the extrema (0.0-1.0)
                                     in the float image. */
  double fmaxpix;    /* float equivalent of max level of converted image */
  double fminpix;    /* float equivalent of min level of converted image */
                                  /* extrema levels found in final image
                                     e.g. 0-255 in 8 bit images. */
  int imaxpix;       /* int equivalent of max level of converted image */ 
  int iminpix;       /* int equivalent of min level of converted image */
  int depth;         /* number bits per pixel 8 -> 256 levels */
} ConvHeader;

PUBLIC char *optarg;               /* used by getopt() */
PUBLIC int optind=0;
PUBLIC int opterr=0;
PUBLIC int optopt=0;

PRIVATE ConvHeader *ParseArg(ConvHeader *,int,char **,FILE **,OUTPTR *);
PRIVATE OUTPUT_IMAGE_INT **ConvertImage(ImageHeader *,ConvHeader *,
                                        RECON_FLOAT **);
PRIVATE void OutputImage(OUTPTR *,ImageHeader *,ConvHeader *,
                         OUTPUT_IMAGE_INT **);
PRIVATE void CalculateStatistics(ImageHeader *,ConvHeader *,RECON_FLOAT **);

PUBLIC BOOLEAN verbose=FALSE;
 
/*****************************************************************************/
/*****************************************************************************/

PUBLIC int main(int argc,char *argv[])

{ /* main() */
  ImageHeader *ih;           /* image header */
  FILE *inptr=NULL;          /* input file pointer */
  OUTPTR out;                /* output file pointer */
  ConvHeader *ch;            /* conversion mode header */
  OUTPUT_IMAGE_INT **intf;   /* integer version of image array */
  RECON_FLOAT **f;           /* input image array */

  ch=(ConvHeader *)calloc(1,sizeof(ConvHeader));       /* parse command line */
  ParseArg(ch,argc,argv,&inptr,&out);

  AllocateImageHeader(ih,NDIM);                      /* set up to read image */

  if(!ioImageHeader(ih,fread,inptr)) {
    errno=HEADER_ERROR;
    perror("recon2img : unable to read image header ");
    return(errno);
  }
  if(verbose)
    ImageHeaderPrintPersonalData(stderr,ih);
  f=(RECON_FLOAT **)mdmalloc(NDIM,sizeof(RECON_FLOAT),ALLOC_DISCRETE,ih->lim);
  if(f<0) {
    perror("recon2img : mdmalloc failure on input array ");
    return(errno);
  }

  if(ih->fsize == sizeof(RECON_FLOAT))                         /* read image */
    mdio(inptr,2,ih->fsize,sizeof(RECON_FLOAT),(void **)f,fread,ih->lim);
  else {
    fprintf(stderr,"recon2img : incompatible float size\n");
    fprintf(stderr,"        file : %1d bytes      program : %1d bytes\n",
            ih->fsize,sizeof(RECON_FLOAT));
    errno = INCOMPATIBLE_DATA;
    return(errno);
  }
                        
  intf=ConvertImage(ih,ch,f);                        /* convert to int array */
  OutputImage(&out,ih,ch,intf);                        /* output final image */

  return(errno);
} /* main() */

/*****************************************************************************/

/* this function converts the image data to an int array with the appropriate
   number of bits per pixel */

PRIVATE OUTPUT_IMAGE_INT **ConvertImage
(
 ImageHeader *ih, /* image header */
 ConvHeader *ch,  /* conversion mode header */
 RECON_FLOAT **f  /* image array */
)
{ /* ConvertImage() */
  double maxf;             /* largest element in image */
  double minf;             /* smallest element in image */
  OUTPUT_IMAGE_INT **intf; /* output int image array */

  { /* determine extrema */
    int i,j;  /* counter */

    maxf=minf=f[LOX][LOY];
    for(i=ih->lim[LOX];i<=ih->lim[HIX];i++)
      for(j=ih->lim[LOY];j<=ih->lim[HIY];j++) {
        maxf=MAX(maxf,f[i][j]);
        minf=MIN(minf,f[i][j]);
      } 

    if(verbose) {
      fprintf(stderr,"recon2img : largest  pixel = %g\n",maxf);
      fprintf(stderr,"recon2img : smallest pixel = %g\n",minf);
    }
  } /* determine extrema */
  
  { /* normalize */
    int i,j; /* counters */
    double a;  /* conversion factor */

                                   /* account for user defined dynamic range */
    minf= ch->fminpix;
    maxf= ch->fmaxpix;

    if(verbose) {
      fprintf(stderr,"recon2img : white has a value of %f\n",maxf);
      fprintf(stderr,"recon2img : black has a value of %f\n",minf);
    }

    a=(ch->imaxpix-ch->iminpix)/(maxf-minf);  /* calculate conversion factor */

    for(i=ih->lim[LOX];i<=ih->lim[HIX];i++)
      for(j=ih->lim[LOY];j<=ih->lim[HIY];j++) {
        if(f[i][j]>maxf) f[i][j]=maxf;
        if(f[i][j]<minf) f[i][j]=minf;
        f[i][j] = a*(f[i][j]-minf)+ch->iminpix;
      }      
  } /* normalize */

  if(verbose)
    CalculateStatistics(ih,ch,f);

  { /* fill up int image array */
    if(sizeof(RECON_FLOAT)==sizeof(OUTPUT_IMAGE_INT))  /* can use same array */
      intf = (OUTPUT_IMAGE_INT **)f;
    else
      intf=(OUTPUT_IMAGE_INT **)mdmalloc(NDIM,sizeof(OUTPUT_IMAGE_INT),
                                         ALLOC_DISCRETE,ih->lim);

    { /* fill up int image array */
      int i,j; /* counters */
      
      for(i=ih->lim[LOX];i<=ih->lim[HIX];i++)
        for(j=ih->lim[LOY];j<=ih->lim[HIY];j++)
          intf[i][j] = NINT(f[i][j]);
    } /* fill up int array */

    /*   if((void *)intf != (void *)f) */       /* free up extraneous array */
    /*  mdfree(NDIM,f,sizeof(RECON_FLOAT),ih->lim); */ /* memory leak */
  } /* fill up int image array */
    
  return(intf);
} /* ConvertImage() */

/*****************************************************************************/

/* this function outputs the image */

PRIVATE void OutputImage
(
 OUTPTR *out,               /* output file pointer */
 ImageHeader *ih,            /* image header */
 ConvHeader *ch,             /* conversion mode header */
 OUTPUT_IMAGE_INT **intf     /* image array */
)
{ /* OutputImage() */

  switch(ch->ftype) {
  case IMAGE_GIF:
    fprintf(stderr,"recon2img : code to support GIF output not yet written\n");
    break;
  case IMAGE_TIFF:
    {
        /* due to the way C defines arrays f[X][Y], Y direction is contiguous.
           or f[col][row] (by columns).  Thus, each row must be copied
           to a secondary array for use by the TIFF library.

           Also, the TIFF library also appears to count the rows from top
           down */
      TIFFSetField(out->tiff,TIFFTAG_IMAGEWIDTH,(uint32)(ih->dim[X]));
      TIFFSetField(out->tiff,TIFFTAG_IMAGELENGTH,(uint32)(ih->dim[Y]));
      TIFFSetField(out->tiff,TIFFTAG_ORIENTATION,ORIENTATION_TOPLEFT);
      TIFFSetField(out->tiff,TIFFTAG_SAMPLESPERPIXEL,1);
      TIFFSetField(out->tiff,TIFFTAG_BITSPERSAMPLE,ch->depth);
      TIFFSetField(out->tiff,TIFFTAG_PLANARCONFIG,PLANARCONFIG_CONTIG);
      TIFFSetField(out->tiff,TIFFTAG_COMPRESSION,ch->ctype);
      TIFFSetField(out->tiff, TIFFTAG_PHOTOMETRIC,PHOTOMETRIC_MINISBLACK);

      switch (ch->ctype) {
      case COMPRESSION_JPEG:
        TIFFSetField(out->tiff, TIFFTAG_JPEGQUALITY,RECON_DEFAULT_JPEGQUALITY);
        TIFFSetField(out->tiff, TIFFTAG_JPEGCOLORMODE,JPEGCOLORMODE_RGB);
        break;
      case COMPRESSION_LZW:
      case COMPRESSION_DEFLATE:
      /*TIFFSetField(out->tiff,TIFFTAG_PREDICTOR,RECON_DEFAULT_LZWPREDICTOR);*/
        perror("recon2img : compression not supported yet ");
        break;
      }

      TIFFSetField(out->tiff,TIFFTAG_ROWSPERSTRIP,
                   TIFFDefaultStripSize(out->tiff,RECON_DEFAULT_ROWSPERSTRIP));

      { /* write data */
        uint32 tiffrow=0;           /* row of tiff file to be written */
        int y;                      /* row in image to be read */
        OUTPUT_IMAGE_INT *row;      /* copied row data */

        row=(OUTPUT_IMAGE_INT *)mdmalloc(1,
                    (sizeof(OUTPUT_IMAGE_INT)*(ih->lim[HIY]-ih->lim[LOY]+1)),
                    ALLOC_CONTIGUOUS,&(ih->lim[LOY]));
        for (tiffrow=0,y=ih->lim[HIY];y>=ih->lim[LOY];tiffrow++,y--) {
          int x;   /* column in image to be read */
          for(x=ih->lim[LOX];x<=ih->lim[HIX];x++) row[x]=intf[x][y];

          if(TIFFWriteScanline(out->tiff,&(row[ih->lim[LOX]]),tiffrow,0)<0)
            break;
        }
        (void)TIFFClose(out->tiff);
      } /* write data */
    }
    break;
  default:
    perror("recon2img : OutputImage : illegal mode ");
    errno = UNSUPPORTED_MODE;
  }
} /* OutputImage() */
  
/*****************************************************************************/

/* this function takes the normalized array and determines the distribution
   of array elements. */

#define MAXBUF   64
#define DATACOL   8
#define DATAROW  MAXBUF/DATACOL

PRIVATE void CalculateStatistics
(
 ImageHeader *ih, /* image header */
 ConvHeader *ch,  /* conversion mode header */
 RECON_FLOAT **f  /* image array */
)
{ /* CalculateStatistics() */
  int i,j;         /* counter */
  int buf[MAXBUF]; /* statistics counter */

  int shift=(ch->imaxpix-ch->iminpix+1)/(2*MAXBUF);         /* bits to shift */

  for(i=0;i<MAXBUF;i++) buf[i]=0;

  for(i=ih->lim[LOX];i<=ih->lim[HIX];i++)
    for(j=ih->lim[LOY];j<=ih->lim[HIY];j++)
      buf[NINT(f[i][j])>>shift]++;

  fprintf(stderr,"statistics: in %d bins\n",MAXBUF);

  for(i=0;i<DATACOL;i++)
    fprintf(stderr," i  count ");
  fprintf(stderr,"\n");

  for(i=0;i<DATAROW;i++) {
    for(j=0;j<DATACOL;j++) {
      int ele = i+j*DATACOL;
      fprintf(stderr,"%2d%7d",ele,buf[ele]);
      if(j<(DATACOL-1)) fprintf(stderr," ");
    }
    fprintf(stderr,"\n");
  }
} /* CalculateStatistics() */
#undef DATAROW
#undef DATACOL
#undef MAXBUF

/*****************************************************************************/

/* this function processes the command line arguments of recon.
   See the definition of char *errormsg[] below for legal options.

   The default output is an 8 bit uncompressed TIFF file */

PRIVATE ConvHeader *ParseArg
(
 ConvHeader *h,      /* image header */
 int argc,           /* argument count */
 char **argv,        /* arguments */
 FILE **inptr,       /* input file pointer */
 OUTPTR *out         /* output file pointer */
)
{ /* ParseArg() */
  char opt;                   /* option */
  char *outfname=NULL;        /* output filename: outfile not opened until all
                                 args parsed */
  char *infname=NULL;         /* filename of input file */
  BOOLEAN error=FALSE;        /* tests if error in arguments */
  CHOICE LimType=UNDEFINED;   /* extrema vs window/level to define limits */ 

  float center=1.0;          /* center of window */
  float width=0.1;            /* width of window */

  char ReconOptions[]="i:o:t:z:h:l:c:w:b:v";
  char *errormsg[] = {
    "recon2img [-i <name>] [-o <name>] [-t <type>] [-c <type>] [-h #] [-l #] [-b #]",
    "   -i <input file name>",
    "   -o <output file name>",
    "   -t <file type: gif || tiff>",
    "   -z <output file compression type: none || jpeg || lzw || zip>",
    "   -h <float equivalent of max pixel : default 1.1>",
    "   -l <float equivalent of min pixel : default 0.9>",
    "   -c <float equivalent of center pixel : default 1.0",
    "   -w <float equivalent of window width : default 0.1", 
    "   -b <number of bits per pixel>",
    "   -v verbose mode",
    NULL
  };

  h->fminpix=0.9;  /* default dynamic range */
  h->fmaxpix=1.1;

  do {
    opt = getopt(argc,argv,ReconOptions);
    switch(opt) {
    case 'i':                                           /* input filename */
      *inptr=fopen(optarg,"r");
      if(*inptr == NULL) {
        perror("recon2img : unable to open input file");
        error = TRUE;
      } else {
        infname=malloc(strlen(optarg)+1);
        strcpy(infname,optarg);
      }
      break;
    case 'o':                                           /* output filename */
      outfname=malloc(strlen(optarg)+1); /* since filetype may not be known, */
      strcpy(outfname,optarg);           /* at this time, file opened after */
      break;                             /* all args parsed  */
    case 'h':  /* float equivalent of max level of converted image */
      if((LimType==UNDEFINED)||(LimType==EXTREMA)) {
        h->fmaxpix = atof(optarg);
        LimType = EXTREMA;
      } else {
        perror("recon2img : can't mix -h and -l with -w -c");
        error = TRUE;
      }
      break;
    case 'l':  /* float equivalent of min level of converted image */
      if((LimType==UNDEFINED)||(LimType==EXTREMA)) {
        h->fminpix = atof(optarg);
        LimType = EXTREMA;
      } else {
        perror("recon2img : can't mix -h and -l with -w -c");
        error = TRUE;
      }
      break;
    case 'c':  /* float equivalent of center pixel */
      if((LimType==UNDEFINED)||(LimType==CENTER)) {
        center = atof(optarg);
        LimType = CENTER;
      } else {
        perror("recon2img : can't mix -h and -l with -w -c");
        error = TRUE;
      }
      break;
    case 'w':  /* float equivalent of window width */
      if((LimType==UNDEFINED)||(LimType==CENTER)) {
        width = atof(optarg);
        LimType = CENTER;
      } else {
        perror("recon2img : can't mix -h and -l with -w -c");
        error = TRUE;
      }
      break;
    case 't':  /* output file type */
      STOUPPER(optarg);
      if (!strcmp(optarg,"TIFF"))
        h->ftype = IMAGE_TIFF;
      else if (!strcmp(optarg,"GIF"))
        h->ftype = IMAGE_GIF;
      else {
        perror("ParseArg : illegal image type");
        error = TRUE;
      }
      break;
    case 'z':  /* compression mode */
      STOUPPER(optarg);

      if (!strcmp(optarg,"NONE"))
        h->ctype = COMPRESSION_NONE;
      else if (!strcmp(optarg,"JPEG"))
        h->ctype = COMPRESSION_JPEG;
      else if (!strcmp(optarg,"LZW"))
        h->ctype = COMPRESSION_LZW;
      else if (!strcmp(optarg,"ZIP"))
        h->ctype = COMPRESSION_DEFLATE;
      else {
        perror("ParseArg : illegal compression type");
        error = TRUE;
      }
      break;
    case 'b':                       /* number of bits per pixel */
      h->depth = atoi(optarg);
      h->iminpix = 0;
      h->imaxpix = NINT(pow(2,h->depth)-1);
      break;
    case 'v':                                                /* verbose mode */
      verbose=TRUE;
      break;
    case ':':                                           /* missing parameter */
      perror("recon2img : missing parameter in option");
      error = TRUE;
      break;
    case '?':                                              /* illegal option */
      perror("recon2img : illegal option");
      error = TRUE;
      break;
    case -1:                                           /* all options parsed */
      { /* all args parsed */
                                       /* default is uncompressed 8 bit TIFF */
        if((h->depth==0)&&(h->ctype==0)&&(h->ftype==0)) {
          h->ftype=IMAGE_TIFF;
          h->ctype=COMPRESSION_NONE;
          h->depth=8;
          h->iminpix = 0;
          h->imaxpix = NINT(pow(2,h->depth)-1);
        }
        if(LimType==CENTER) {
          h->fminpix = center-width/2.0;
          h->fmaxpix = center+width/2.0;
        }
        if(h->depth==0) {
          fprintf(stderr,"recon2img : bits per pixel not set\n");
          error = TRUE;
        }
        if(h->ctype==0) {
          fprintf(stderr,"recon2img : compression type not set\n");
          error = TRUE;
        }
        if(h->fmaxpix<=h->fminpix) {
          fprintf(stderr,"recon2img : illegal max/min level: dynamic range\n");
          error = TRUE;
        }
        if(*inptr==NULL) {
          fprintf(stderr,"recon2img : no input file\n");
          error = TRUE;
        }
        if(h->ftype==0) {
          fprintf(stderr,"recon2img : output image file type not set\n");
          error = TRUE;
        } else {
          if(outfname!=NULL) {
            if(h->ftype==IMAGE_TIFF)
              out->tiff=TIFFOpen(outfname,"w");
            else 
              out->ptr = fopen(optarg,"w+");
            if(out->ptr == NULL) { /* NOTE: out->ptr & out->tiff is a union */
              perror("recon2img : unable to open output file");
              error = TRUE;
            } 
          }
        }
        
        if(verbose && !error) {
          if(infname) {
            fprintf(stderr,"recon2img : input filename : %s\n",infname);
            free(infname);
          } else
            fprintf(stderr,"recon2img : input from stdin\n");

          if(outfname) {
            fprintf(stderr,"recon2img : output filename : %s\n",outfname);
            free(outfname);
          } else
            fprintf(stderr,"recon2img : output to stdout\n");

          switch(h->ftype) {
          case IMAGE_TIFF:
            fprintf(stderr,"recon2img : TIFF file output\n");
            break;
          case IMAGE_GIF:
            fprintf(stderr,"recon2img : GIF file output\n");
            break;
          default:
            perror("ParseArg : illegal image type");
            error = TRUE;
          }

          switch(h->ctype) {
          case COMPRESSION_NONE:
            fprintf(stderr,"recon2img : no compression\n");
            break;
          case COMPRESSION_JPEG:
            fprintf(stderr,"recon2img : jpeg compression\n");
            break;
          case COMPRESSION_LZW:
            fprintf(stderr,"recon2img : lzw compression\n");
            break;
          case COMPRESSION_DEFLATE:
            fprintf(stderr,"recon2img : zip compression\n");
            break;
          default:
            perror("recon2img : unknown compression mode");
            error = TRUE;
          }

          if (h->imaxpix<0) {
            fprintf(stderr,"recon2img : illegal # of bits pixel\n");
            fprintf(stderr,"            using default\n");
            h->imaxpix=NINT(pow(2,DEFAULT_BITSPERPIXEL)-1);
            h->depth=DEFAULT_BITSPERPIXEL;
          } else
            fprintf(stderr,"recon2img : %d bits per pixel\n",h->depth);
        }

        if(error)
          USAGE(errormsg);
      } /* all args parsed */
      break;
    default :
      errno = ILLEGAL_ARGUMENTS;
      exit(ILLEGAL_ARGUMENTS);
    }
  } while(opt != -1);

  return(h);

} /* ParseArg() */
